/*
 * GnpinpoutGenerator.h
 *
 *  Created on: Jan 4, 2011
 *      Author: robert goerke
 */

#include <time.h>
#include <getopt.h>
#include <math.h>
#include <cstdlib>
#include <vector>
#include <exception>
#include <iostream>
#include <fstream>
#include <utility>                   // for std::pair
#include <algorithm>                 // for std::for_each
#include <string>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/version.hpp>
#include <boost/random.hpp>
#include <boost/generator_iterator.hpp>
#include <boost/lexical_cast.hpp>

#ifndef GNPINPOUTGENERATOR_H_
#define GNPINPOUTGENERATOR_H_

#define MAX(a,b) ((a>b) ? a:b)

using namespace boost;

typedef std::pair<int, int> Edge;
typedef std::vector<Edge> Edge_vector;

class GnpinpoutGenerator {

public:
	GnpinpoutGenerator();
	GnpinpoutGenerator(int random_add);
	//	virtual ~GnpinpoutGenerator() {};
	void generate(int par_n, int k, double skew_exponent, bool randomize_k, double pout, double pin_global, std::string pin_in_file_name, std::string cluster_sizes_in_file_name, bool par_weighted,
			bool par_verbosity, bool par_report);
	void store_graph_in_file(std::string graph_out_file_name, std::string file_mode);
	//
	//	create a typedef for the Graph type
	//	typedef adjacency_list<listS, vecS, bidirectionalS, no_property, property<edge_weight_t, double> > GraphW;
	//	typedef adjacency_list<vecS, vecS, bidirectionalS> Graph;
	template<typename Graph>
	Graph* get_boost_graph();
	std::vector<std::pair<int, int> > get_edge_vector();
	int get_n();
	int get_m();
	std::vector<double> get_weight_vector();
	std::vector<int> get_cluster_sizes();

private:
	int n;
	std::vector<int> cluster_sizes;
	Edge_vector edge_vector;
	std::vector<double> weight_vector;
	bool verbosity, report;
	bool weighted;
	boost::mt19937 random_gen; //generator

	void read_cluster_sizes(std::string& cluster_sizes_in_file_name, bool& use_cluster_sizes_in_file);
	void read_pins(std::vector<double>& pin, double pin_global, std::string pin_in_file_name, bool& use_pin_in_file);
	template<class T>
	void create_intra_edges(long long& total_intra_edge_count, T& zero_one_random, std::vector<double>& pin);
	template<class T>
	void create_inter_edges(long long& total_inter_edge_count, T& zero_one_random, double pout, std::vector<long long>& cluster_out_edge_cands);
};

GnpinpoutGenerator::GnpinpoutGenerator() {
	random_gen.seed(static_cast<unsigned int> (time(0)));
}

GnpinpoutGenerator::GnpinpoutGenerator(int random_add) {
	random_gen.seed(static_cast<unsigned int> (random_add + time(0)));
}

std::vector<std::pair<int, int> > GnpinpoutGenerator::get_edge_vector() {
	return edge_vector;
}

int GnpinpoutGenerator::get_n() {
	return n;
}

int GnpinpoutGenerator::get_m() {
	return edge_vector.size();
}

std::vector<double> GnpinpoutGenerator::get_weight_vector() {
	return weight_vector;
}

std::vector<int> GnpinpoutGenerator::get_cluster_sizes() {
	return cluster_sizes;
}

template<typename Graph>
Graph* GnpinpoutGenerator::get_boost_graph() {
	if (verbosity || report)
		std::cout << "creating a boost graph\n";
	Graph *gP = NULL;
	if (weighted) {
		gP = new Graph(edge_vector.begin(), edge_vector.end(), weight_vector.begin(), n);
	} else {
		gP = new Graph(edge_vector.begin(), edge_vector.end(), n);
	}

	return gP;
}

void GnpinpoutGenerator::store_graph_in_file(std::string graph_out_file_name, std::string file_mode) {
	//	std::cout<<file_mode<<"\n";
	if ((file_mode != "a") && (file_mode != "b")) {
		return;
	}
	int k = cluster_sizes.size();
	//standard filename
	if (graph_out_file_name == "") {
		if (verbosity || report)
			std::cout << "no file name given, using timestamp \n";
		time_t rawtime;
		struct tm * timeinfo;
		char date_time_str[19];
		time(&rawtime);
		timeinfo = localtime(&rawtime);
		strftime(date_time_str, 20, "%Y-%m-%d_%H-%M-%S", timeinfo);
		graph_out_file_name = std::string("G_n_pin_pout_") + date_time_str;
	}
	std::string ref_clust_out_file_name = graph_out_file_name;
	ref_clust_out_file_name.append("_clusters");

	if (file_mode == "a") {
		graph_out_file_name.append(".txt");
		//		ref_clust_out_file_name.append(".txt");
	} else {
		graph_out_file_name.append(".binary");
		//		ref_clust_out_file_name.append(".binary");
	}
	ref_clust_out_file_name.append(".txt");

	int m = edge_vector.size();
	std::ofstream graph_out_file;
	std::ofstream ref_clust_out_file;
	if (file_mode == "a") {
		graph_out_file.open(graph_out_file_name.c_str(), std::ios::out | std::ios::trunc);
		//		ref_clust_out_file.open(ref_clust_out_file_name.c_str(), std::ios::out | std::ios::trunc);
	} else {
		graph_out_file.open(graph_out_file_name.c_str(), std::ios::out | std::ios::trunc | std::ios::binary);
		//		ref_clust_out_file.open(ref_clust_out_file_name.c_str(), std::ios::out | std::ios::trunc | std::ios::binary);
	}
	ref_clust_out_file.open(ref_clust_out_file_name.c_str(), std::ios::out | std::ios::trunc);

	if (verbosity || report) {
		std::cout << "creating output graph file with mode = " << file_mode << "\n";
		std::cout << "output graph file name = " << graph_out_file_name << "\n";
		std::cout << "output reference clustering file name = " << ref_clust_out_file_name << "\n";
	}

	if (file_mode == "a") { //clear text file
		graph_out_file << n << " " << m;
		if (weighted)
			graph_out_file << " w\n";
		else
			graph_out_file << "\n";

		std::vector<double>::iterator weight_iterator;
		std::vector<Edge>::iterator edge_iterator;
		if (weighted)
			weight_iterator = weight_vector.begin();
		for (edge_iterator = edge_vector.begin(); edge_iterator != edge_vector.end(); ++edge_iterator) {
			graph_out_file << edge_iterator->first << " " << edge_iterator->second;
			if (weighted) {
				graph_out_file << " " << *weight_iterator << "\n";
				++weight_iterator;
			} else {
				graph_out_file << "\n";
			}
		}
	}
	if (file_mode == "b") { //binary file
		graph_out_file.seekp(0, std::ios::beg);
		std::ifstream::pos_type one_int_size = sizeof(int);
		std::ifstream::pos_type one_bool_size = sizeof(bool);
		graph_out_file.write((char*) &n, one_int_size);
		graph_out_file.write((char*) &m, one_int_size);
		graph_out_file.write((char*) &weighted, one_bool_size);

		if (!weighted) { //quick and gescheit
			std::ifstream::pos_type one_edge_size = sizeof(Edge);
			graph_out_file.write((char *) (&edge_vector.front()), m * one_edge_size);
		} else { //gammlig and via loop, polymorphic edge struct tomorrow
			std::ifstream::pos_type one_double_size = sizeof(double);
			std::vector<double>::iterator weight_iterator = weight_vector.begin();
			for (std::vector<Edge>::iterator edge_iterator = edge_vector.begin(); edge_iterator != edge_vector.end(); ++edge_iterator) {
				//						graph_out_file.write((char *)(&weight_vector.front()), m * one_double_size);
				graph_out_file.write((char*) &edge_iterator->first, one_int_size);
				graph_out_file.write((char*) &edge_iterator->second, one_int_size);
				graph_out_file.write((char*) &(*weight_iterator), one_double_size);
				++weight_iterator;
			}
		}
	}
	graph_out_file.close();

	file_mode = "a";
	//	std::cout << "now storing reference clustering (cluster sizes) in " << ref_clust_out_file_name << "\n";
	if (file_mode == "a") { //text file
		ref_clust_out_file << n << "\n";
		for (int i = 0; i < k; ++i) {
			for (int j = 0; j < cluster_sizes[i]; ++j) {
				ref_clust_out_file << i << "\n";
			}
		}
	}
	if (file_mode == "b") { //binary file
		ref_clust_out_file.write((char *) &n, sizeof(int));
		//			ref_clust_out_file.write((char *)(&cluster_sizes.front()), k * sizeof(int));
		for (int i = 0; i < k; ++i) {
			for (int j = 0; j < cluster_sizes[i]; ++j) {
				ref_clust_out_file.write((char *) &i, sizeof(int));
			}
		}
	}
	ref_clust_out_file.close();
}

template<class T>
void GnpinpoutGenerator::create_inter_edges(long long& total_inter_edge_count, T& zero_one_random, double pout, std::vector<long long>& cluster_out_edge_cands) {
	int k = cluster_sizes.size();
	double r;
	int source_node_index = 0, target_node_index = 0;
	double weight = 1.0;
	int cumul_seen_nodes = 0, current_out_cluster = 0;
	long long waiting_time, current_edge_index = -1;
	int source_node_offset = 0, source_node_rel_pos = 0;
	int target_node_offset = cluster_sizes[0], target_node_rel_pos = 0;
	if (pout > 0 && pout <= 1.0) {
		//	int past_c_cumul_c_out_edges = 0;
		cumul_seen_nodes = cluster_sizes[current_out_cluster]; //includes active source candidates
		if (verbosity)
			std::cout << "out_cluster " << current_out_cluster << "\n";
		//    while ((current_edge_index + past_c_cumul_c_out_edges) < cumul_cluster_out_edge_cands[k-1]){
		while (!((current_out_cluster == k - 1) && (current_edge_index >= cluster_out_edge_cands[k - 1]))) { //while not at the end of last cluster
			r = zero_one_random();
			waiting_time = 1 + floor(log(1 - r) / log(1 - pout));
			if (verbosity)
				std::cout << "waiting time " << waiting_time << "\n";
			current_edge_index += waiting_time;
			while ((current_edge_index > cluster_out_edge_cands[current_out_cluster] - 1) && (current_out_cluster < k - 1)) { //while source beyond current clusters -> next cluster
				//   			past_c_cumul_c_out_edges = cumul_cluster_out_edge_cands[current_out_cluster];
				current_edge_index -= cluster_out_edge_cands[current_out_cluster];
				++current_out_cluster;
				source_node_offset = cumul_seen_nodes; //past
				cumul_seen_nodes += cluster_sizes[current_out_cluster];
				target_node_offset = cumul_seen_nodes; //now
				if (verbosity)
					std::cout << "out_cluster " << current_out_cluster << "\n";
			}
			if (current_edge_index < cluster_out_edge_cands[current_out_cluster]) { //if source inside this cluster
				source_node_rel_pos = (current_edge_index) / (n - cumul_seen_nodes); //div for source
				source_node_index = source_node_offset + source_node_rel_pos;
				target_node_rel_pos = current_edge_index % (n - cumul_seen_nodes); //mod for target
				target_node_index = target_node_offset + target_node_rel_pos;
				if (weighted)
					weight = zero_one_random();
				if (verbosity)
					std::cout << "inserting inter-edge " << current_edge_index << ", which is " << source_node_index << " --> " << target_node_index << " w=" << weight << "\n";
				edge_vector.push_back(Edge(source_node_index, target_node_index));
				if (weighted)
					weight_vector.push_back(weight);
				//				}
				++total_inter_edge_count;
			} else {
				if (verbosity)
					std::cout << "too much, end: " << current_edge_index << "\n";
			}
		}
	}
}

template<class T>
void GnpinpoutGenerator::create_intra_edges(long long& total_intra_edge_count, T& zero_one_random, std::vector<double>& pin) {
	double r;
	int source_node_index = 0;
	int target_node_index = 0;
	double weight = 1.0;
	int current_cluster_size;
	int cumul_seen_nodes = 0;
	int k = cluster_sizes.size();
	for (int i = 0; i < k; i++) {
		if (verbosity) {
			std::cout << "dealing with intra cluster " << i << ", it has " << cluster_sizes[i] << " nodes \n";
		}
		if (pin[i] <= 0 || pin[i] > 1.0) {
			cumul_seen_nodes += cluster_sizes[i];
			continue;
		}
		current_cluster_size = cluster_sizes[i];
		long long target_pos = 1, source_pos = -1;
		while (target_pos < current_cluster_size) {
			r = zero_one_random();
			source_pos = source_pos + 1 + (int) floor((log(1 - r)) / (log(1 - pin[i])));
			while ((source_pos >= target_pos) && (target_pos < current_cluster_size)) {
				source_pos -= target_pos;
				++target_pos;
			}
			if (target_pos < current_cluster_size) {
				source_node_index = source_pos + cumul_seen_nodes;
				target_node_index = target_pos + cumul_seen_nodes;
				if (weighted)
					weight = zero_one_random();
				if (verbosity)
					std::cout << "creating edge " << source_node_index << " --> " << target_node_index << " w=" << weight << " inside cluster " << i;
				edge_vector.push_back(Edge(source_node_index, target_node_index));
				if (weighted)
					weight_vector.push_back(weight);
				//				}
				if (verbosity)
					std::cout << " done\n";
				++total_intra_edge_count;
			}
		}
		cumul_seen_nodes += current_cluster_size;
	}
}

void GnpinpoutGenerator::read_cluster_sizes(std::string& cluster_sizes_in_file_name, bool& use_cluster_sizes_in_file) {
	int new_n = 0;
	double sum = 0;
	int k = cluster_sizes.size();
	std::vector<double> rel_cluster_sizes;
	double a_relative_nk;

	//reading some numbers from file, that need not make much sense
	std::ifstream cluster_sizes_in_file(cluster_sizes_in_file_name.c_str(), std::ios::in);
	if (cluster_sizes_in_file.is_open()) {
		if (verbosity) {
			std::cout << "reading: ";
		}
		while (!cluster_sizes_in_file.eof()) {
			std::string a_line;
			std::getline(cluster_sizes_in_file, a_line);
			if (a_line == "") {
				continue;
			}
			if (verbosity) {
				std::cout << a_line << " ";
			}
			a_relative_nk = boost::lexical_cast<double>(a_line);
			if (a_relative_nk <= 0) {
				if (verbosity) {
					std::cout << "...ignored ";
				}
				continue;
			}
			rel_cluster_sizes.push_back(a_relative_nk);
			sum += a_relative_nk;
			++k;
		}
		if (verbosity) {
			std::cout << " found " << k << " good values \n" << "sizes are: ";
		}

		//reading finished, now normalizing and writing into actual vector
		cluster_sizes.reserve(k);
		for (unsigned int i = 0; i < rel_cluster_sizes.size(); ++i) {
			cluster_sizes[i] = floor(0.5 + n * rel_cluster_sizes[i] / sum);
			new_n += cluster_sizes[i];
			if (verbosity) {
				std::cout << cluster_sizes[i] << " ";
			}
		}
		if (verbosity) {
			std::cout << "\n";
		}
		if (n != new_n) {
			std::cout << " warning, n changed to " << new_n << " due to crooked values in your input file\n";
			n = new_n;
		}
	} else {
		std::cout << "could not open cluster_size-file, falling back to normal mode...";
		use_cluster_sizes_in_file = false;
	}
}

void GnpinpoutGenerator::read_pins(std::vector<double>& pin, double pin_global, std::string pin_in_file_name, bool& use_pin_in_file) {
	double a_pin;
	int i = 0;
	int k = cluster_sizes.size();
	//reading some numbers from file, that need not make much sense
	std::ifstream pin_in_file(pin_in_file_name.c_str(), std::ios::in);
	if (pin_in_file.is_open()) {
		if (verbosity) {
			std::cout << "reading: ";
		}
		while (!pin_in_file.eof() && i < k) {
			std::string a_line;
			std::getline(pin_in_file, a_line);
			if (a_line == "") {
				continue;
			}
			if (verbosity) {
				std::cout << a_line << " ";
			}
			a_pin = boost::lexical_cast<double>(a_line);
			if (a_pin <= 0 || a_pin > 1.0) {
				if (verbosity) {
					std::cout << "...set to pin_global=" << pin_global << " ";
				}
				a_pin = pin_global;
			}
			pin[i++] = a_pin;
		}
		while (i < k) {
			pin[i++] = pin_global;
			if (verbosity) {
				std::cout << "...missing set to pin_global=" << pin_global << " ";
			}
		}
		if (verbosity) {
			std::cout << "\n";
		}
	} else {
		std::cout << "could not open pin-file, falling back to normal mode...";
		use_pin_in_file = false;
	}
}

void GnpinpoutGenerator::generate(int par_n, int k, double skew_exponent, bool randomize_k, double pout, double pin_global, std::string pin_in_file_name, std::string cluster_sizes_in_file_name,
		bool par_weighted, bool par_verbosity, bool par_report) {
	//	  std::cout << sizeof(long long) << "\n";

	//nicening and checking arguments and declaring and resetting helper variables
	cluster_sizes.clear();
	edge_vector.clear();
	weight_vector.clear();
	weighted = par_weighted;
	n = par_n;
	verbosity = par_verbosity && par_report;
	report = par_report;
	bool use_cluster_sizes_in_file = false;
	bool use_pin_in_file = false;
	std::vector<double> pin;
	k = std::min(MAX(k,0), n);
	use_cluster_sizes_in_file = (cluster_sizes_in_file_name != "");
	pout = std::min(std::max(pout, 0.0), 1.0);
	pin_global = std::min(std::max(pin_global, 0.0), 1.0);
	use_pin_in_file = (pin_in_file_name != "");

	//initialize random generator
	//	boost::mt19937 random_gen; //generator
	boost::uniform_01<> uniform_random_unit_interval;
	boost::variate_generator<boost::mt19937&, boost::uniform_01<> > zero_one_random(random_gen, uniform_random_unit_interval);
	//start clock
	clock_t start_clock_time = clock();

	//possibly randomize k
	if (randomize_k && !use_cluster_sizes_in_file) {
		boost::uniform_int<> uniform_random_int_interval(1, floor(log(n) * sqrt(n)));
		boost::variate_generator<boost::mt19937&, boost::uniform_int<> > cluster_count_random_gen(random_gen, uniform_random_int_interval);
		k = cluster_count_random_gen();
	}

	//report on stuff to be used
	if (verbosity || report) {
		std::cout << "using the following parameters:" << "\n" << " \t-n number of nodes = " << n << "\n";
		if (randomize_k && !use_cluster_sizes_in_file) {
			std::cout << " \t-r randomization of number of clusters used, yields " << k << "\n";
		} else if (use_cluster_sizes_in_file) {
			std::cout << " \t-c acquiring cluster sizes and count from file " << cluster_sizes_in_file_name << "\n";
		} else {//no random or otherwisely obtained k
			std::cout << " \t-k fixed number of clusters = " << k << "\n";
		}
		if (use_pin_in_file) {
			std::cout << " \t-p individual internal edge probs to be used from file " << pin_in_file_name << "\n";
		} else {
			std::cout << " \t-i internal edge probability is " << pin_global << "\n";
		}
		if (skew_exponent != 1.0) {
			std::cout << " \t-s cluster sizes skewed by exponent " << skew_exponent << "\n";
		}
		std::cout << " \t-x external edge probability = " << pout << "\n";
		//				  << " \t-c boost graph constructor: batch or iterative, default is " << edge_adding_mode << "\n"
		std::cout << " \t-w weighing edges is switched to " << weighted << "\n";
		if (verbosity) {
			std::cout << " \t-v being verbose\n";
		}
		std::cout << "using boost lib version: " << BOOST_LIB_VERSION << std::endl;
	}

	//if necessary, read cluster sizes, overrides given value of k, reserve pin's size k
	if (use_cluster_sizes_in_file) {
		read_cluster_sizes(cluster_sizes_in_file_name, use_cluster_sizes_in_file);
		k = cluster_sizes.size();
	}
	pin.clear();
	pin.reserve(k);

	// if necessary generate partition and pins:
	// somehow random cluster sizes, if not set otherwise
	int c_index;
	if (!use_cluster_sizes_in_file) {
		for (int i = 0; i < k; ++i) {
			cluster_sizes.push_back(0);
		}
		//    boost::uniform_int<> uniform_random_cluster_int(0,(k-1));
		//    boost::variate_generator< boost::mt19937&, boost::uniform_int<> > uniform_random_cluster(random_gen, uniform_random_cluster_int);
		for (int i = 0; i < n; ++i) {
			//		c_index = uniform_random_cluster();
			c_index = floor(pow(zero_one_random(), skew_exponent) * k);
			if (c_index >= k || c_index < 0)
				std::cout << "wrong cluster access";
			++cluster_sizes[c_index];
		}
	} else { //use the set values from in_file!
		//		all done above...
	}

	// boring pins, if not set otherwise
	if (use_pin_in_file) {
		read_pins(pin, pin_global, pin_in_file_name, use_pin_in_file);
	}
	//maybe it didn't work ?
	if (!use_pin_in_file) {
		for (int i = 0; i < k; ++i) {
			pin[i] = pin_global;
		}
	}

	if (verbosity) {
		std::cout << "pins are: ";
		for (int i = 0; i < k; ++i) {
			std::cout << pin[i] << " ";
		}
		std::cout << "\n";
	}

	double expected_total_intra_edge_count = 0;
	for (int i = 0; i < k; ++i) {
		if (verbosity)
			std::cout << "cluster: " << i << ", size is: " << cluster_sizes[i] << "\n";
		expected_total_intra_edge_count += (pin[i] * ((cluster_sizes[i] / 2.0) * (cluster_sizes[i] - 1)));
		//		std::cout << "cluster sizes: " << cluster_sizes[i] << " pairs " << ((cluster_sizes[i] / 2.0) * (cluster_sizes[i]-1)) << " " << sizeof(long) << "\n";
		//		std::cout << "e i e: " << expected_total_intra_edge_count << "\n";
	}

	//compute offsets of edge numbering for candidate inter-edges
	long long current_c_out_edge_cands;
	int cumul_seen_nodes = 0;
	std::vector<long long> cluster_out_edge_cands;
	//	std::vector<int> cumul_cluster_out_edge_cands;
	double expected_total_inter_edge_count = 0;
	for (int i = 0; i < k; ++i) {
		cumul_seen_nodes += cluster_sizes[i];
		current_c_out_edge_cands = (long long) cluster_sizes[i] * (n - cumul_seen_nodes);
		cluster_out_edge_cands.push_back(current_c_out_edge_cands);
		//    	if (i>0)
		//    		cumul_cluster_out_edge_cands.push_back(cumul_cluster_out_edge_cands[i-1] + current_c_out_edge_cands);
		//    	else
		//    		cumul_cluster_out_edge_cands.push_back(current_c_out_edge_cands);
		if (verbosity)
			std::cout << "edge-candidates outgoing from: " << i << ": " << current_c_out_edge_cands << ".\n";
		expected_total_inter_edge_count += (current_c_out_edge_cands * pout);
	}

	//compute expected size of edge vector, declare and initialize
	size_t edgeSet_size = floor((expected_total_intra_edge_count + expected_total_inter_edge_count) * (double) 1.05);
	edge_vector.clear();
	weight_vector.clear(); //aligned to edge_vector
	edge_vector.reserve(edgeSet_size);
	if (weighted)
		weight_vector.reserve(edgeSet_size);

	//counting intra and inter edges
	long long total_intra_edge_count = 0;
	long long total_inter_edge_count = 0;

	//intra edges:
	create_intra_edges(total_intra_edge_count, zero_one_random, pin);

	//inter-edges:
	create_inter_edges(total_inter_edge_count, zero_one_random, pout, cluster_out_edge_cands);

	//brief reporting:
	int m = total_inter_edge_count + total_intra_edge_count;
	if (report)
		std::cout << "total edges:\t " << m << "\t (intra: " << total_intra_edge_count << ",\t inter: " << total_inter_edge_count << ")\n";
	if (report)
		std::cout << "expected:\t " << +round(expected_total_inter_edge_count + expected_total_intra_edge_count) << "\t (intra: " << round(expected_total_intra_edge_count) << ",\t inter: " << round(
				expected_total_inter_edge_count) << ")\n";
	clock_t end_clock_time = clock();
	if (report)
		std::cout << "time consumed:\t " << (end_clock_time - start_clock_time) / CLOCKS_PER_SEC << "s\n";

	//end
	if (report)
		std::cout << "generator finished. " << "\n";

	//	if (gP == NULL)
	//		delete gP;
	//	if (gwP == NULL)
	//		delete gwP;

	return;
}

#endif /* GNPINPOUTGENERATOR_H_ */
